@@ -30,7 +30,7 @@ typedef guchar *(*PixopsLineFunc) (int *weights, int n_x, int n_y,
 typedef void (*PixopsPixelFunc) (guchar *dest, int dest_x, int dest_channels, int dest_has_alpha,
 				 int src_has_alpha, int check_size, guint32 color1,
 				 guint32 color2,
-				 int r, int g, int b, int a);
+				 guint r, guint g, guint b, guint a);
 
 static int
 get_check_shift (int check_size)
@@ -175,37 +175,43 @@ pixops_composite_nearest (guchar        *dest_buf,
 	  else
 	    a0 = overall_alpha;
 
-	  if (dest_has_alpha)
-	    {
-	      unsigned int w0 = 0xff * a0;
-	      unsigned int w1 = (0xff - a0) * dest[3];
-	      unsigned int w = w0 + w1;
+          switch (a0)
+            {
+            case 0:
+              break;
+            case 255:
+              dest[0] = p[0];
+              dest[1] = p[1];
+              dest[2] = p[2];
+              if (dest_has_alpha)
+                dest[3] = 0xff;
+              break;
+            default:
+              if (dest_has_alpha)
+                {
+                  unsigned int w0 = 0xff * a0;
+                  unsigned int w1 = (0xff - a0) * dest[3];
+                  unsigned int w = w0 + w1;
 
-	      if (w != 0)
-		{
 		  dest[0] = (w0 * p[0] + w1 * dest[0]) / w;
 		  dest[1] = (w0 * p[1] + w1 * dest[1]) / w;
 		  dest[2] = (w0 * p[2] + w1 * dest[2]) / w;
 		  dest[3] = w / 0xff;
-		}
-	      else
-		{
-		  dest[0] = 0;
-		  dest[1] = 0;
-		  dest[2] = 0;
-		  dest[3] = 0;
-		}
-	    }
-	  else
-	    {
-	      dest[0] = (a0 * p[0] + (0xff - a0) * dest[0]) / 0xff;
-	      dest[1] = (a0 * p[1] + (0xff - a0) * dest[1]) / 0xff;
-	      dest[2] = (a0 * p[2] + (0xff - a0) * dest[2]) / 0xff;
-	      
-	      if (dest_channels == 4)
-		dest[3] = 0xff;
-	    }
+                }
+              else
+                {
+                  unsigned int a1 = 0xff - a0;
+		  unsigned int tmp;
 
+		  tmp = a0 * p[0] + a1 * dest[0] + 0x80;
+                  dest[0] = (tmp + (tmp >> 8)) >> 8;
+		  tmp = a0 * p[1] + a1 * dest[1] + 0x80;
+                  dest[1] = (tmp + (tmp >> 8)) >> 8;
+		  tmp = a0 * p[2] + a1 * dest[2] + 0x80;
+                  dest[2] = (tmp + (tmp >> 8)) >> 8;
+                }
+              break;
+            }
 	  dest += dest_channels;
 	  x += x_step;
 	}
@@ -282,31 +288,48 @@ pixops_composite_color_nearest (guchar        *dest_buf,
 	  else
 	    a0 = overall_alpha;
 
-	  if (a0 == 255)
-	    {
+          switch (a0)
+            {
+            case 0:
+              if (((j + check_x) >> check_shift) & 1)
+                {
+                  dest[0] = r2; 
+                  dest[1] = g2; 
+                  dest[2] = b2;
+                }
+              else
+                {
+                  dest[0] = r1; 
+                  dest[1] = g1; 
+                  dest[2] = b1;
+                }
+            break;
+            case 255:
 	      dest[0] = p[0];
 	      dest[1] = p[1];
 	      dest[2] = p[2];
-	    }
-	  else
-	    if (((j + check_x) >> check_shift) & 1)
-	      {
-		tmp = ((int) p[0] - r2) * a0;
-		dest[0] = r2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-		tmp = ((int) p[1] - g2) * a0;
-		dest[1] = g2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-		tmp = ((int) p[2] - b2) * a0;
-		dest[2] = b2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-	      }
-	    else
-	      {
-		tmp = ((int) p[0] - r1) * a0;
-		dest[0] = r1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-		tmp = ((int) p[1] - g1) * a0;
-		dest[1] = g1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-		tmp = ((int) p[2] - b1) * a0;
-		dest[2] = b1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
-	      }
+              break;
+            default:
+              if (((j + check_x) >> check_shift) & 1)
+                {
+                  tmp = ((int) p[0] - r2) * a0;
+                  dest[0] = r2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                  tmp = ((int) p[1] - g2) * a0;
+                  dest[1] = g2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                  tmp = ((int) p[2] - b2) * a0;
+                  dest[2] = b2 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                }
+              else
+                {
+                  tmp = ((int) p[0] - r1) * a0;
+                  dest[0] = r1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                  tmp = ((int) p[1] - g1) * a0;
+                  dest[1] = g1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                  tmp = ((int) p[2] - b1) * a0;
+                  dest[2] = b1 + ((tmp + (tmp >> 8) + 0x80) >> 8);
+                }
+              break;
+            }
 	  
 	  if (dest_channels == 4)
 	    dest[3] = 0xff;
@@ -320,7 +343,7 @@ pixops_composite_color_nearest (guchar        *dest_buf,
 static void
 composite_pixel (guchar *dest, int dest_x, int dest_channels, int dest_has_alpha,
 		 int src_has_alpha, int check_size, guint32 color1, guint32 color2,
-		 int r, int g, int b, int a)
+		 guint r, guint g, guint b, guint a)
 {
   if (dest_has_alpha)
     {
@@ -368,7 +391,7 @@ composite_line (int *weights, int n_x, int n_y,
       int *pixel_weights;
       
       pixel_weights = weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK) * n_x * n_y;
-
+      
       for (i=0; i<n_y; i++)
 	{
 	  guchar *q = src[i] + x_scaled * src_channels;
@@ -524,7 +547,7 @@ composite_line_22_4a4_mmx_stub (int *weights, int n_x, int n_y,
 static void
 composite_pixel_color (guchar *dest, int dest_x, int dest_channels, int dest_has_alpha,
 		       int src_has_alpha, int check_size, guint32 color1, guint32 color2,
-		       int r, int g, int b, int a)
+		       guint r, guint g, guint b, guint a)
 {
   int dest_r, dest_g, dest_b;
   int check_shift = get_check_shift (check_size);
@@ -670,7 +693,7 @@ composite_line_color_22_4a4_mmx_stub (int *weights, int n_x, int n_y,
 static void
 scale_pixel (guchar *dest, int dest_x, int dest_channels, int dest_has_alpha,
 	     int src_has_alpha, int check_size, guint32 color1, guint32 color2,
-	     int r, int g, int b, int a)
+	     guint r, guint g, guint b, guint a)
 {
   if (src_has_alpha)
     {
@@ -819,7 +842,7 @@ scale_line_22_33_mmx_stub (int *weights, int n_x, int n_y,
 
 static guchar *
 scale_line_22_33 (int *weights, int n_x, int n_y,
-		  guchar *dest, guchar *dest_end, int dest_channels, int dest_has_alpha,
+		  guchar *dest, int dest_x, guchar *dest_end, int dest_channels, int dest_has_alpha,
 		  guchar **src, int src_channels, gboolean src_has_alpha,
 		  int x_init, int x_step, int src_width,
 		  int check_size, guint32 color1, guint32 color2)
@@ -839,8 +862,8 @@ scale_line_22_33 (int *weights, int n_x, int n_y,
       q0 = src0 + x_scaled * 3;
       q1 = src1 + x_scaled * 3;
       
-      pixel_weights = (int *)((char *)weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS - 4)) & (SUBSAMPLE_MASK << 4)));
-      
+      pixel_weights = weights + ((x >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK) * 4;
+
       w1 = pixel_weights[0];
       w2 = pixel_weights[1];
       w3 = pixel_weights[2];
@@ -862,9 +885,9 @@ scale_line_22_33 (int *weights, int n_x, int n_y,
       g += w4 * q1[5];
       b += w4 * q1[6];
 
-      dest[0] = r >> 16;
-      dest[1] = g >> 16;
-      dest[2] = b >> 16;
+      dest[0] = (r + 0x8000) >> 16;
+      dest[1] = (g + 0x8000) >> 16;
+      dest[2] = (b + 0x8000) >> 16;
       
       dest += 3;
       
@@ -943,21 +966,36 @@ pixops_process (guchar         *dest_buf,
 		PixopsPixelFunc pixel_func)
 {
   int i, j;
-  int x, y;
+  int x, y;			/* X and Y position in source (fixed_point) */
   guchar **line_bufs = g_new (guchar *, filter->n_y);
 
-  int x_step = (1 << SCALE_SHIFT) / scale_x;
-  int y_step = (1 << SCALE_SHIFT) / scale_y;
+  int x_step = (1 << SCALE_SHIFT) / scale_x; /* X step in source (fixed point) */
+  int y_step = (1 << SCALE_SHIFT) / scale_y; /* Y step in source (fixed point) */
+
+  int check_shift = check_size ? get_check_shift (check_size) : 0;
 
-  int dest_x;
   int scaled_x_offset = floor (filter->x_offset * (1 << SCALE_SHIFT));
 
-  int run_end_index = (((src_width - filter->n_x + 1) << SCALE_SHIFT) - scaled_x_offset - 1) / x_step + 1 - render_x0;
-  int check_shift = check_size ? get_check_shift (check_size) : 0;
+  /* Compute the index where we run off the end of the source buffer. The furthest
+   * source pixel we access at index i is:
+   *
+   *  ((render_x0 + i) * x_step + scaled_x_offset) >> SCALE_SHIFT + filter->n_x - 1
+   *
+   * So, run_end_index is the smallest i for which this pixel is src_width, i.e, for which:
+   *
+   *  (i + render_x0) * x_step >= ((src_width - filter->n_x + 1) << SCALE_SHIFT) - scaled_x_offset
+   *
+   */
+#define MYDIV(a,b) ((a) > 0 ? (a) / (b) : ((a) - (b) + 1) / (b))    /* Division so that -1/5 = -1 */
+  
+  int run_end_x = (((src_width - filter->n_x + 1) << SCALE_SHIFT) - scaled_x_offset);
+  int run_end_index = MYDIV (run_end_x + x_step - 1, x_step) - render_x0;
+  run_end_index = MIN (run_end_index, render_x1 - render_x0);
 
   y = render_y0 * y_step + floor (filter->y_offset * (1 << SCALE_SHIFT));
   for (i = 0; i < (render_y1 - render_y0); i++)
     {
+      int dest_x;
       int y_start = y >> SCALE_SHIFT;
       int x_start;
       int *run_weights = filter->weights + ((y >> (SCALE_SHIFT - SUBSAMPLE_BITS)) & SUBSAMPLE_MASK) * filter->n_x * filter->n_y * SUBSAMPLE;
@@ -1008,12 +1046,12 @@ pixops_process (guchar         *dest_buf,
 	  outbuf += dest_channels;
 	}
 
-      new_outbuf = (*line_func)(run_weights, filter->n_x, filter->n_y,
-				outbuf, dest_x,
-				MIN (outbuf_end, dest_buf + dest_rowstride * i + run_end_index * dest_channels),
-				dest_channels, dest_has_alpha,
-				line_bufs, src_channels, src_has_alpha,
-				x, x_step, src_width, check_size, tcolor1, tcolor2);
+      new_outbuf = (*line_func) (run_weights, filter->n_x, filter->n_y,
+				 outbuf, dest_x,
+				 dest_buf + dest_rowstride * i + run_end_index * dest_channels,
+				 dest_channels, dest_has_alpha,
+				 line_bufs, src_channels, src_has_alpha,
+				 x, x_step, src_width, check_size, tcolor1, tcolor2);
 
       dest_x += (new_outbuf - outbuf) / dest_channels;
 
@@ -1057,8 +1095,8 @@ tile_make_weights (PixopsFilter *filter, double x_scale, double y_scale, double
     for (j_offset=0; j_offset<SUBSAMPLE; j_offset++)
       {
 	int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
-	double x = (double)j_offset / 16;
-	double y = (double)i_offset / 16;
+	double x = (double)j_offset / SUBSAMPLE;
+	double y = (double)i_offset / SUBSAMPLE;
 	int i,j;
 	  
 	for (i = 0; i < n_y; i++)
@@ -1143,8 +1181,8 @@ bilinear_make_fast_weights (PixopsFilter *filter, double x_scale, double y_scale
     for (j_offset=0; j_offset<SUBSAMPLE; j_offset++)
       {
 	int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
-	double x = (double)j_offset / 16;
-	double y = (double)i_offset / 16;
+	double x = (double)j_offset / SUBSAMPLE;
+	double y = (double)i_offset / SUBSAMPLE;
 	int i,j;
 
 	if (x_scale > 1.0)	/* Bilinear */
@@ -1156,6 +1194,10 @@ bilinear_make_fast_weights (PixopsFilter *filter, double x_scale, double y_scale
 	  }
 	else			/* Tile */
 	  {
+	    /*           x
+	     * ---------|--.-|----|--.-|-------  SRC
+	     * ------------|---------|---------  DEST
+	     */
 	    for (i = 0; i < n_x; i++)
 	      {
 		if (i < x)
@@ -1184,6 +1226,10 @@ bilinear_make_fast_weights (PixopsFilter *filter, double x_scale, double y_scale
 	  }
 	else			/* Tile */
 	  {
+	    /*           y
+	     * ---------|--.-|----|--.-|-------  SRC
+	     * ------------|---------|---------  DEST
+	     */
 	    for (i = 0; i < n_y; i++)
 	      {
 		if (i < y)
@@ -1205,7 +1251,7 @@ bilinear_make_fast_weights (PixopsFilter *filter, double x_scale, double y_scale
 
 	for (i = 0; i < n_y; i++)
 	  for (j = 0; j < n_x; j++)
-	    *(pixel_weights + n_x * i + j) = 65536 * x_weights[j] * x_scale * y_weights[i] * y_scale * overall_alpha;
+	    *(pixel_weights + n_x * i + j) = 65536 * x_weights[j] * x_scale * y_weights[i] * y_scale * overall_alpha + 0.5;
       }
 
   g_free (x_weights);
@@ -1287,8 +1333,8 @@ bilinear_make_weights (PixopsFilter *filter, double x_scale, double y_scale, dou
     for (j_offset=0; j_offset<SUBSAMPLE; j_offset++)
       {
 	int *pixel_weights = filter->weights + ((i_offset*SUBSAMPLE) + j_offset) * n_x * n_y;
-	double x = (double)j_offset / 16;
-	double y = (double)i_offset / 16;
+	double x = (double)j_offset / SUBSAMPLE;
+	double y = (double)i_offset / SUBSAMPLE;
 	int i,j;
 	  
 	for (i = 0; i < n_y; i++)
@@ -1323,7 +1369,7 @@ pixops_composite_color (guchar         *dest_buf,
 			gboolean        src_has_alpha,
 			double          scale_x,
 			double          scale_y,
-			GdkInterpType   interp_type,
+			PixopsInterpType   interp_type,
 			int             overall_alpha,
 			int             check_x,
 			int             check_y,
@@ -1352,7 +1398,7 @@ pixops_composite_color (guchar         *dest_buf,
 
   switch (interp_type)
     {
-    case GDK_INTERP_NEAREST:
+    case PIXOPS_INTERP_NEAREST:
       pixops_composite_color_nearest (dest_buf, render_x0, render_y0, render_x1, render_y1,
 				      dest_rowstride, dest_channels, dest_has_alpha,
 				      src_buf, src_width, src_height, src_rowstride, src_channels, src_has_alpha,
@@ -1360,15 +1406,15 @@ pixops_composite_color (guchar         *dest_buf,
 				      check_x, check_y, check_size, color1, color2);
       return;
 
-    case GDK_INTERP_TILES:
+    case PIXOPS_INTERP_TILES:
       tile_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
       
-    case GDK_INTERP_BILINEAR:
+    case PIXOPS_INTERP_BILINEAR:
       bilinear_make_fast_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
       
-    case GDK_INTERP_HYPER:
+    case PIXOPS_INTERP_HYPER:
       bilinear_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
     }
@@ -1390,6 +1436,30 @@ pixops_composite_color (guchar         *dest_buf,
   g_free (filter.weights);
 }
 
+/**
+ * pixops_composite:
+ * @dest_buf: pointer to location to store result
+ * @render_x0: x0 of region of scaled source to store into @dest_buf
+ * @render_y0: y0 of region of scaled source to store into @dest_buf
+ * @render_x1: x1 of region of scaled source to store into @dest_buf
+ * @render_y1: x1 of region of scaled source to store into @dest_buf
+ * @dest_rowstride: rowstride of @dest_buf
+ * @dest_channels: number of channels in @dest_buf
+ * @dest_has_alpha: whether @dest_buf has alpha
+ * @src_buf: pointer to source pixels
+ * @src_width: width of source (used for clipping)
+ * @src_height: height of source (used for clipping)
+ * @src_rowstride: rowstride of source
+ * @src_channels: number of channels in @src_buf
+ * @src_has_alpha: whether @src_buf has alpha
+ * @scale_x: amount to scale source by in X direction
+ * @scale_y: amount to scale source by in Y direction
+ * @interp_type: type of enumeration
+ * @overall_alpha: overall alpha factor to multiply source by
+ * 
+ * Scale source buffer by scale_x / scale_y, then composite a given rectangle
+ * of the result into the destination buffer.
+ **/
 void
 pixops_composite (guchar        *dest_buf,
 		  int            render_x0,
@@ -1407,7 +1477,7 @@ pixops_composite (guchar        *dest_buf,
 		  gboolean       src_has_alpha,
 		  double         scale_x,
 		  double         scale_y,
-		  GdkInterpType  interp_type,
+		  PixopsInterpType  interp_type,
 		  int            overall_alpha)
 {
   PixopsFilter filter;
@@ -1431,22 +1501,22 @@ pixops_composite (guchar        *dest_buf,
 
   switch (interp_type)
     {
-    case GDK_INTERP_NEAREST:
+    case PIXOPS_INTERP_NEAREST:
       pixops_composite_nearest (dest_buf, render_x0, render_y0, render_x1, render_y1,
 				dest_rowstride, dest_channels, dest_has_alpha,
 				src_buf, src_width, src_height, src_rowstride, src_channels,
 				src_has_alpha, scale_x, scale_y, overall_alpha);
       return;
 
-    case GDK_INTERP_TILES:
+    case PIXOPS_INTERP_TILES:
       tile_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
       
-    case GDK_INTERP_BILINEAR:
+    case PIXOPS_INTERP_BILINEAR:
       bilinear_make_fast_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
       
-    case GDK_INTERP_HYPER:
+    case PIXOPS_INTERP_HYPER:
       bilinear_make_weights (&filter, scale_x, scale_y, overall_alpha / 255.);
       break;
     }
@@ -1490,7 +1560,7 @@ pixops_scale (guchar        *dest_buf,
 	      gboolean       src_has_alpha,
 	      double         scale_x,
 	      double         scale_y,
-	      GdkInterpType  interp_type)
+	      PixopsInterpType  interp_type)
 {
   PixopsFilter filter;
   PixopsLineFunc line_func;
@@ -1508,32 +1578,36 @@ pixops_scale (guchar        *dest_buf,
 
   switch (interp_type)
     {
-    case GDK_INTERP_NEAREST:
+    case PIXOPS_INTERP_NEAREST:
       pixops_scale_nearest (dest_buf, render_x0, render_y0, render_x1, render_y1,
 			    dest_rowstride, dest_channels, dest_has_alpha,
 			    src_buf, src_width, src_height, src_rowstride, src_channels, src_has_alpha,
 			    scale_x, scale_y);
       return;
 
-    case GDK_INTERP_TILES:
+    case PIXOPS_INTERP_TILES:
       tile_make_weights (&filter, scale_x, scale_y, 1.0);
       break;
       
-    case GDK_INTERP_BILINEAR:
+    case PIXOPS_INTERP_BILINEAR:
       bilinear_make_fast_weights (&filter, scale_x, scale_y, 1.0);
       break;
       
-    case GDK_INTERP_HYPER:
+    case PIXOPS_INTERP_HYPER:
       bilinear_make_weights (&filter, scale_x, scale_y, 1.0);
       break;
     }
 
+  if (filter.n_x == 2 && filter.n_y == 2 && dest_channels == 3 && src_channels == 3)
+    {
 #ifdef USE_MMX
-  if (filter.n_x == 2 && filter.n_y == 2 &&
-      found_mmx && dest_channels == 3 && src_channels == 3)
-    line_func = scale_line_22_33_mmx_stub;
+      if (found_mmx)
+	line_func = scale_line_22_33_mmx_stub;
+      else
+#endif
+	line_func = scale_line_22_33;
+    }
   else
-#endif    
     line_func = scale_line;
   
   pixops_process (dest_buf, render_x0, render_y0, render_x1, render_y1,
